/*******************************************************************************
nfs.c

Copyright (C) 2000 Simon Nieuviarts

An implementaion of the NFS protocol.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version. 

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*******************************************************************************/
/*
TODO:
- Replace some OSKIT_FAILED( err ) by err, suppress some or add some others.
- Do tests on memory allocation, verify the reference counters on the objects.
- Set locking everywhere.
- What to do with the problem of conversion from 64-bit to 32-bit file offsets.
*/

#include <oskit/dev/osenv_mem.h>
#include <oskit/dev/osenv_timer.h>
#include <oskit/fs/filesystem.h>
#include <oskit/fs/dir.h>
#include <oskit/io/bufio.h>
#include <oskit/io/absio.h>
#include <oskit/net/socket.h>
#include <oskit/freebsd/stdio.h>
#include <oskit/freebsd/stdarg.h>
#include <oskit/freebsd/stdlib.h>
#include <oskit/freebsd/string.h>
#include "rpc.h"
#include "nfs.h"


#define NFS_PROG 100003
#define NFS_VERSION 2
#define MOUNT_PROG 100005
#define MOUNT_VERSION 1
#define MNTPROC_MNT 1
#define NFSPROC_GETATTR 1
#define NFSPROC_SETATTR 2
#define NFSPROC_LOOKUP 4
#define NFSPROC_READLINK 5
#define NFSPROC_READ 6
#define NFSPROC_WRITE 8
#define NFSPROC_CREATE 9
#define NFSPROC_REMOVE 10
#define NFSPROC_RENAME 11
#define NFSPROC_LINK 12
#define NFSPROC_SYMLINK 13
#define NFSPROC_MKDIR 14
#define NFSPROC_RMDIR 15
#define NFSPROC_READDIR 16
#define NFSPROC_STATFS 17
#define MNTPATHLEN 1024
#define NFS_MAXNAMLEN 255
#define NFS_MAXPATHLEN 1024


/* Activates extra verifications. */
#define MEM_DEBUG
#undef DEBUG


typedef char nfs_fhandle[32] ;


typedef enum nfs_ftype
{
    NFS_NFNON,
    NFS_NFREG,
    NFS_NFDIR,
    NFS_NFBLK,
    NFS_NFCHR,
    NFS_NFLNK
} nfs_ftype ;


struct timeval
{
    oskit_u32_t tv_sec ;
    oskit_u32_t tv_usec ;
} ;


typedef struct nfs_fattr
{
    nfs_ftype type ;
    unsigned int mode ;
    unsigned int nlink ;
    unsigned int uid ;
    unsigned int gid ;
    unsigned int size ;
    unsigned int blocksize ;
    unsigned int rdev ;
    unsigned int blocks ;
    unsigned int fsid ;
    unsigned int fileid ;
    struct timeval atime ;
    struct timeval mtime ;
    struct timeval ctime ;
} nfs_fattr ;


typedef struct nfs_sattr
{
    unsigned int mode ;
    unsigned int uid ;
    unsigned int gid ;
    unsigned int size ;
    struct timeval atime ;
    struct timeval mtime ;
} nfs_sattr ;


static struct oskit_file_ops nfs_file_ops ;
static struct oskit_dir_ops nfs_dir_ops ;
static struct oskit_filesystem_ops nfs_fsys_ops ;
static struct oskit_bufio_ops nfs_bufio_ops ;


typedef struct nfs_fsys
{
    oskit_filesystem_t fsysi ;
    char *hostname ;
    char *dirname ;
    oskit_u32_t ref_counter ;
    oskit_osenv_mem_t *mem ;
    oskit_rpc_t *rpc ;
    oskit_dir_t *root;
} nfs_fsys_t ;


typedef struct nfs_file
{
    oskit_file_t filei ;
    oskit_osenv_mem_t *mem ;
    char *name ;
    nfs_fhandle fh ;
    oskit_u32_t ref_counter ;
    nfs_fsys_t *fs ;
    oskit_bufio_t bufioi ;  /* Operations for a file. */
    oskit_dir_t *parent ;   /* Parent directory for a directory. */
} nfs_file_t ;


/*******************************************************************************
Debug code.
*******************************************************************************/


#ifdef DEBUG

#define LOG_BUF_SZ (1024*1024)

#else

#define LOG_BUF_SZ 0

#endif


static char log_buffer[LOG_BUF_SZ] ;
static char *log_ptr = log_buffer ;
static oskit_size_t log_free = LOG_BUF_SZ ;
static char log_stopped = 0 ;


#ifdef DEBUG

static void
do_log( const char *fmt, ... )
{
    va_list lst ;
    oskit_size_t len ;

    if ( log_stopped ) return ;
    va_start( lst, fmt ) ;
    len = vsnprintf( log_ptr, log_free, fmt, lst ) ;

printf( "%s", log_ptr ) ;

    if ( len > log_free ) len = log_free ;
    log_ptr += len ;
    log_free -= len ;
    va_end( lst ) ;
}

#else

#define do_log(...) ((void)0)

#endif


/* Stop NFS debug log. Must be done before copying the log in an NFS file. */
void
stop_nfs_log( void )
{
    log_stopped = 1 ;
}


/* Restart NFS debug log. Can be done after copying the log in an NFS file. */
void
restart_nfs_log( void )
{
    log_stopped = 0 ;
    log_ptr = log_buffer ;
    log_free = LOG_BUF_SZ ;
}


/* Read at most COUNT bytes from the NFS debug log in BUF. */
oskit_size_t
read_nfs_log( char *buf, oskit_size_t count )
{
    static char *my_log_ptr = log_buffer ;
    oskit_size_t log_len ;

    log_len = log_ptr - my_log_ptr ;
    if ( log_len < count ) count = log_len ;
    memcpy( buf, my_log_ptr, count ) ;
    my_log_ptr += count ;
    return count ;
}


/*******************************************************************************
Code to debug memory allocations.
*******************************************************************************/


#ifdef MEM_DEBUG

static oskit_size_t allocated = 0 ;


static void *OSKIT_COMCALL
OSKIT_OSENV_MEM_ALLOC( oskit_osenv_mem_t *o, oskit_size_t size,
                       osenv_memflags_t flags, unsigned int align )
{
    void *ptr ;

    ptr = oskit_osenv_mem_alloc( o, size, flags, align ) ;

    if ( ptr == 0 )
        do_log( "Memory allocation of %d bytes failed. Total allocated %d.\n",
                size, allocated ) ;
    else
    {
        allocated += size ;
        do_log( "Memory allocation of %d bytes at %#x. Total allocated %d.\n",
                size, (oskit_addr_t)ptr, allocated ) ;
    }

    return ptr ;
}


static OSKIT_COMDECL_V
OSKIT_OSENV_MEM_FREE( oskit_osenv_mem_t *o, void *block,
                      osenv_memflags_t flags, oskit_size_t size )
{
    allocated -= size ;

    do_log( "Memory deallocation of %d bytes at %#x. Total allocated %d.\n",
            size, (oskit_addr_t)block, allocated ) ;
    oskit_osenv_mem_free( o, block, flags, size ) ;
}

#else

#define OSKIT_OSENV_MEM_ALLOC oskit_osenv_mem_alloc
#define OSKIT_OSENV_MEM_FREE oskit_osenv_mem_free

#endif


/*******************************************************************************
Error code conversions.
*******************************************************************************/


#define NFS_OK                  0               /* v2 v3 */
#define NFSERR_PERM             1               /* v2 v3 */
#define NFSERR_NOENT            2               /* v2 v3 */
#define NFSERR_IO               5               /* v2 v3 */
#define NFSERR_NXIO             6               /* v2 v3 */
#define NFSERR_ACCES            13              /* v2 v3 */
#define NFSERR_EXIST            17              /* v2 v3 */
#define NFSERR_NODEV            19              /* v2 v3 */
#define NFSERR_NOTDIR           20              /* v2 v3 */
#define NFSERR_ISDIR            21              /* v2 v3 */
#define NFSERR_FBIG             27              /* v2 v3 */
#define NFSERR_NOSPC            28              /* v2 v3 */
#define NFSERR_ROFS             30              /* v2 v3 */
#define NFSERR_NAMETOOLONG      63              /* v2 v3 */
#define NFSERR_NOTEMPTY         66              /* v2 v3 */
#define NFSERR_DQUOT            69              /* v2 v3 */
#define NFSERR_STALE            70              /* v2 v3 */
#define NFSERR_WFLUSH           99              /* v2    */


static struct
{
    int nfs_err ;
    oskit_error_t err ;

} nfs_errtbl[] =
{
    { NFS_OK,               0                  },
    { NFSERR_PERM,          OSKIT_EPERM        },
    { NFSERR_NOENT,         OSKIT_ENOENT       },
    { NFSERR_IO,            OSKIT_EIO          },
    { NFSERR_NXIO,          OSKIT_ENXIO        },
    { NFSERR_ACCES,         OSKIT_EACCES       },
    { NFSERR_EXIST,         OSKIT_EEXIST       },
    { NFSERR_NODEV,         OSKIT_ENODEV       },
    { NFSERR_NOTDIR,        OSKIT_ENOTDIR      },
    { NFSERR_ISDIR,         OSKIT_EISDIR       },
    { NFSERR_FBIG,          OSKIT_EFBIG        },
    { NFSERR_NOSPC,         OSKIT_ENOSPC       },
    { NFSERR_ROFS,          OSKIT_EROFS        },
    { NFSERR_NAMETOOLONG,   OSKIT_ENAMETOOLONG },
    { NFSERR_NOTEMPTY,      OSKIT_ENOTEMPTY    },
    { NFSERR_DQUOT,         OSKIT_EDQUOT       },
    { NFSERR_STALE,         OSKIT_ESTALE       },
    { -1,                   OSKIT_EIO          }
} ;


static oskit_error_t
nfs_to_oskit_error( oskit_u32_t nfs_err )
{
    int i = 0 ;

    while( nfs_errtbl[i].nfs_err != -1 )
    {
        if ( nfs_errtbl[i].nfs_err == nfs_err )
            return nfs_errtbl[i].err ;
        i++ ;
    }
    return nfs_errtbl[i].err ;
}


/*******************************************************************************
oskit_dir_t and oskit_file_t interfaces.
*******************************************************************************/


static OSKIT_COMDECL
nfs_dir_query( oskit_dir_t *dir, const oskit_iid_t *iid, void **out_ihandle )
{
    nfs_file_t *file = (nfs_file_t *)dir ;

    if ( memcmp( iid, &oskit_iunknown_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_posixio_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_file_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_dir_iid, sizeof(*iid) ) == 0)
    {
        *out_ihandle = file ;
        file->ref_counter++ ;
        return 0 ;
    }

    *out_ihandle = 0 ;
    return OSKIT_E_NOINTERFACE ;
}


static OSKIT_COMDECL
nfs_file_query( oskit_file_t *file, const oskit_iid_t *iid, void **out_ihandle )
{
    nfs_file_t *f = (nfs_file_t *)file ;

    if ( memcmp( iid, &oskit_iunknown_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_posixio_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_file_iid, sizeof(*iid) ) == 0 )
    {
        *out_ihandle = f ;
        f->ref_counter++ ;
        return 0 ;
    }

    if ( memcmp( iid, &oskit_bufio_iid, sizeof(*iid) ) == 0 ||
	 memcmp( iid, &oskit_absio_iid, sizeof(*iid) ) == 0 )
    {
        *out_ihandle = &f->bufioi ;
        f->ref_counter++ ;
        return 0 ;
    }

    *out_ihandle = 0 ;
    return OSKIT_E_NOINTERFACE;
}


static OSKIT_COMDECL_U
nfs_file_addref( oskit_file_t *file )
{
    return ++((nfs_file_t *)file)->ref_counter ;
}


static OSKIT_COMDECL_U nfs_fsys_release( oskit_filesystem_t *fs ) ;


static OSKIT_COMDECL_U
nfs_file_release( oskit_file_t *file )
{
    nfs_file_t *f = (nfs_file_t *)file ;
    oskit_osenv_mem_t *mem = f->mem ;

    if ( --f->ref_counter ) return f->ref_counter ;

    nfs_fsys_release( (oskit_filesystem_t *)f->fs ) ;
    OSKIT_OSENV_MEM_FREE( mem, f->name,
                          OSENV_NONBLOCKING, strlen(f->name) + 1 ) ;
    if ( f->filei.ops == (struct oskit_file_ops *)&nfs_dir_ops )
    {
        /* This is a directory. */
        if ( f->parent )
            nfs_file_release( (oskit_file_t *)f->parent ) ;
    }
    else
    {
        /* This is a file. */
    }

    OSKIT_OSENV_MEM_FREE( mem, f, OSENV_NONBLOCKING, sizeof(*f) ) ;
    oskit_osenv_mem_release( mem ) ;
    return 0 ;
}


/* No reference must be added on the file... */
static oskit_error_t
nfsproc_getattr( nfs_file_t *f, nfs_fattr *attr )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_rpc_t *rpc = f->fs->rpc ;

    oskit_rpc_call_init( rpc, NFSPROC_GETATTR ) ;
    oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        int i ;
        oskit_u32_t *fa = (oskit_u32_t *)attr ;

        if ( oskit_rpc_reply_len( rpc ) != sizeof(nfs_fattr) )
            return OSKIT_E_FAIL ;

        for( i=0; i<(sizeof(nfs_fattr)/4); i++ )
        {
            err = oskit_rpc_reply_rxdr_int( rpc, fa ) ;
            if ( OSKIT_FAILED( err ) ) return err ;
            fa++ ;
        }

        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


static OSKIT_COMDECL
nfs_file_stat( oskit_file_t *file, struct oskit_stat *st )
{
    oskit_error_t err ;
    nfs_fattr fa ;

    err = nfsproc_getattr( (nfs_file_t *)file, &fa ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    st->dev = 0 ;
    st->ino = fa.fileid ;
    st->mode = fa.mode ;
    st->nlink = fa.nlink ;
    st->uid = fa.uid ;
    st->gid = fa.gid ;
    st->rdev = fa.rdev ;
    st->atime.tv_sec = fa.atime.tv_sec ;
    st->atime.tv_nsec = 1000 * fa.atime.tv_usec ;
    st->mtime.tv_sec = fa.mtime.tv_sec ;
    st->mtime.tv_nsec = 1000 * fa.mtime.tv_usec ;
    st->ctime.tv_sec = fa.ctime.tv_sec ;
    st->ctime.tv_nsec = 1000 * fa.ctime.tv_usec ;
    st->size = fa.size ;
    st->blocks = fa.blocks ;
    st->blksize = fa.blocksize ;

    return 0;
}


/* Write a sattr structure. */
static oskit_error_t
nfs_call_wxdr_sattr( oskit_rpc_t *rpc, nfs_sattr *attr )
{
    oskit_error_t err ;
    int i ;
    int *fa = (int *)attr ;

    for( i=0; i<(sizeof(nfs_sattr)/4); i++ )
    {
        err = oskit_rpc_call_wxdr_int( rpc, *fa ) ;
        if ( OSKIT_FAILED( err ) ) return err ;
        fa++ ;
    }
    return 0 ;
}


/* No reference must be added on the file... */
static OSKIT_COMDECL
nfsproc_setattr( nfs_file_t *f, nfs_sattr *attr )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_rpc_t *rpc = f->fs->rpc ;

    oskit_rpc_call_init( rpc, NFSPROC_SETATTR ) ;
    oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;
    nfs_call_wxdr_sattr( rpc, attr ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        if ( oskit_rpc_reply_len( rpc ) != sizeof(nfs_fattr) )
            return OSKIT_E_FAIL ;
        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


static OSKIT_COMDECL
nfs_file_setstat( oskit_file_t *file, oskit_u32_t mask,
                  const struct oskit_stat *stats )
{
    nfs_sattr attr ;

    if ( ( mask & OSKIT_STAT_MODE ) )
        attr.mode = stats->mode ;
    else
        attr.mode = -1 ;

    if ( ( mask & OSKIT_STAT_UID ) )
        attr.uid = stats->uid ;
    else
        attr.uid = -1 ;

    if ( ( mask & OSKIT_STAT_GID ) )
        attr.gid = stats->gid ;
    else
        attr.gid = -1 ;

    if ( ( mask & OSKIT_STAT_SIZE ) )
        attr.size = stats->size ;
    else
        attr.size = -1 ;

    if ( ( mask & OSKIT_STAT_ATIME ) )
    {
        attr.atime.tv_sec = stats->atime.tv_sec ;
        attr.atime.tv_usec = stats->atime.tv_nsec / 1000 ;
    }
    else
    {
        attr.atime.tv_sec = -1 ;
        attr.atime.tv_usec = -1 ;
    }

    if ( ( mask & OSKIT_STAT_MTIME ) )
    {
        attr.mtime.tv_sec = stats->mtime.tv_sec ;
        attr.mtime.tv_usec = stats->mtime.tv_nsec / 1000 ;
    }
    else
    {
        attr.mtime.tv_sec = -1 ;
        attr.mtime.tv_usec = -1 ;
    }

    return nfsproc_setattr( (nfs_file_t *)file, &attr ) ;
}


static OSKIT_COMDECL
no_pathconf( oskit_file_t *file, oskit_s32_t option, oskit_s32_t *out_val )
{
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_file_sync( oskit_file_t *file, oskit_bool_t wait )
{
    /* The data is already sent to the server. */
    return 0 ;
}


static OSKIT_COMDECL
nfs_file_datasync( oskit_file_t *file, oskit_bool_t wait )
{
    /* The data is already sent to the server. */
    return 0 ;
}


static OSKIT_COMDECL
nfs_file_access( oskit_file_t *file, oskit_amode_t mask )
{
    /* XXX To change ??? */
    return 0 ;
}


static OSKIT_COMDECL
nfs_file_readlink( oskit_file_t *file, char *buf, oskit_u32_t len,
                   oskit_u32_t *out_actual )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    nfs_file_t *f = (nfs_file_t *)file ;
    oskit_rpc_t *rpc = f->fs->rpc ;

    oskit_rpc_call_init( rpc, NFSPROC_READLINK ) ;
    oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        char *p ;
        oskit_u32_t l ;

        err = oskit_rpc_reply_rxdr_opaque( rpc, &p, &l ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        if ( l > len ) l = len ;
        memcpy( buf, p, l ) ;
        if ( l < len ) buf[l] = 0 ;

        *out_actual = l ;
        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


static OSKIT_COMDECL
nfs_file_open( oskit_file_t *file, oskit_oflags_t flags,
               struct oskit_openfile **out_openfile )
{
    /* Functionnality not supported by this filesystem. */
    *out_openfile = 0 ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_dir_open( oskit_dir_t *dir, oskit_oflags_t flags,
              struct oskit_openfile **out_opendir )
{
    if ( (flags & OSKIT_O_ACCMODE) != OSKIT_O_RDONLY )
        return OSKIT_EISDIR ;

    /* Functionnality not supported by this filesystem. */
    *out_opendir = 0 ;
    return 0 ;
}


/* References must be added on FS and PARENT before call. NAME and FH are copied
(no pointer are kept on them). */
static nfs_file_t *
nfs_file_allocate( nfs_fsys_t *fs, const char *name, nfs_fhandle *fh,
                   void *ops, oskit_dir_t *parent )
{
    nfs_file_t *file ;
    oskit_size_t len ;
    char *n ;

    len = strlen( name ) + 1 ;

    file = OSKIT_OSENV_MEM_ALLOC( fs->mem, sizeof(nfs_file_t),
                                  OSENV_NONBLOCKING, 0 ) ;
    if ( !file ) return 0 ;

    n = OSKIT_OSENV_MEM_ALLOC( fs->mem, len, OSENV_NONBLOCKING, 0 ) ;
    if ( !n )
    {
        OSKIT_OSENV_MEM_FREE( fs->mem, file, OSENV_NONBLOCKING, sizeof(*file) ) ;
        return 0 ;
    }

    file->filei.ops = ops ;
    oskit_osenv_mem_addref( fs->mem ) ;
    file->mem = fs->mem ;
    memcpy( n, name, len ) ;
    file->name = n ;
    memcpy( &file->fh, fh, sizeof(nfs_fhandle) ) ;
    file->ref_counter = 1 ;
    file->fs = fs ;
    if ( ops == &nfs_dir_ops )
        file->parent = parent ;
    else
        file->bufioi.ops = &nfs_bufio_ops ;

    return file ;
}


static OSKIT_COMDECL
nfs_dir_lookup( oskit_dir_t *dir, const char *name, oskit_file_t **out_file )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;
    nfs_file_t *d = (nfs_file_t *)dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    if ( ( l == 0 ) || ( ( l == 1 ) && ( name[0] == '.' ) ) )
    {
        *out_file = (oskit_file_t *)d ;
        return 0 ;
    }

    if ( ( l == 2 ) && ( name[0] == '.' ) && ( name[1] == '.' ) )
    {
        if ( d->parent )
            *out_file = (oskit_file_t *)(d->parent) ;
        else
            *out_file = (oskit_file_t *)d ;
        return 0 ;
    }

    oskit_rpc_call_init( rpc, NFSPROC_LOOKUP ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_string( rpc, name ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        nfs_fhandle *fh ;
        nfs_file_t *file ;
        void *ops ;

        if ( oskit_rpc_reply_len( rpc ) !=
             sizeof(nfs_fhandle) + sizeof(nfs_fattr) )
            return OSKIT_E_FAIL ;

        oskit_rpc_reply_rxdr_data( rpc, (void**)&fh, sizeof(nfs_fhandle ) ) ;
        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        if ( res == NFS_NFDIR )
        {
            d->ref_counter++ ;
            ops = &nfs_dir_ops ;
        }
        else
        {
            dir = 0 ;
            ops = &nfs_file_ops ;
        }

        d->fs->ref_counter++ ;
        file = nfs_file_allocate( d->fs, name, fh, ops, dir ) ;
        *out_file = (oskit_file_t *)file ;
        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


static OSKIT_COMDECL
nfs_dir_create( oskit_dir_t *dir, const char *name,
                oskit_bool_t excl, oskit_mode_t mode,
                oskit_file_t **out_file )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;
    nfs_file_t *d = (nfs_file_t *)dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;
    nfs_sattr sa ;

    if ( excl )
    {
        oskit_file_t *f ;

        err = oskit_dir_lookup( dir, name, &f ) ;
        if ( !OSKIT_FAILED( err ) ) return OSKIT_EEXIST ;
        if ( err != OSKIT_ENOENT ) return err ;
    }

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_CREATE ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, name, l ) ;
    sa.mode = mode ;
    sa.uid = 0 ;
    sa.gid = 0 ;
    sa.size = 0 ;
    sa.atime.tv_sec = -1 ;
    sa.atime.tv_usec = -1 ;
    sa.mtime.tv_sec = -1 ;
    sa.mtime.tv_usec = -1 ;
    nfs_call_wxdr_sattr( rpc, &sa ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        nfs_fhandle *fh ;
        nfs_file_t *file ;

        if ( oskit_rpc_reply_len( rpc ) !=
             sizeof(nfs_fhandle) + sizeof(nfs_fattr) )
            return OSKIT_E_FAIL ;

        err = oskit_rpc_reply_rxdr_data( rpc, (void**)&fh,
                                         sizeof(nfs_fhandle ) ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        d->fs->ref_counter++ ;
        file = nfs_file_allocate( d->fs, name, fh, &nfs_file_ops, 0 ) ;
        *out_file = (oskit_file_t *)file ;
        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


static OSKIT_COMDECL
nfs_dir_link( oskit_dir_t *dir, const char *name, oskit_file_t *file )
{
    nfs_file_t *d = (nfs_file_t *) dir ;
    nfs_file_t *f = (nfs_file_t *) file ;
    oskit_rpc_t *rpc = d->fs->rpc ;
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_LINK ) ;
    oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, name, l ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    return nfs_to_oskit_error( res ) ;
}


static OSKIT_COMDECL
nfs_dir_unlink( oskit_dir_t *dir, const char *name )
{
    nfs_file_t *d = (nfs_file_t *) dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_REMOVE ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, name, l ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    return nfs_to_oskit_error( res ) ;
}


static OSKIT_COMDECL
nfs_dir_rename( oskit_dir_t *old_dir, const char *old_name,
                oskit_dir_t *new_dir, const char *new_name )
{
    nfs_file_t *od = (nfs_file_t *) old_dir ;
    nfs_file_t *nd = (nfs_file_t *) new_dir ;
    oskit_rpc_t *rpc = od->fs->rpc ;
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t ol ;
    oskit_size_t nl ;

    ol = strlen( old_name ) ;
    if ( ol > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;
    nl = strlen( new_name ) ;
    if ( nl > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_RENAME ) ;
    oskit_rpc_call_wxdr_data( rpc, od->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, old_name, ol ) ;
    oskit_rpc_call_wxdr_data( rpc, nd->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, new_name, nl ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    return nfs_to_oskit_error( res ) ;
}


static OSKIT_COMDECL
nfs_dir_mkdir( oskit_dir_t *dir, const char *name, oskit_mode_t mode )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;
    nfs_file_t *d = (nfs_file_t *)dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;
    nfs_sattr sa ;

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_MKDIR ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, name, l ) ;
    sa.mode = mode ;
    sa.uid = 0 ;
    sa.gid = 0 ;
    sa.size = 0 ;
    sa.atime.tv_sec = -1 ;
    sa.atime.tv_usec = -1 ;
    sa.mtime.tv_sec = -1 ;
    sa.mtime.tv_usec = -1 ;
    nfs_call_wxdr_sattr( rpc, &sa ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    return nfs_to_oskit_error( res ) ;
}


static OSKIT_COMDECL
nfs_dir_rmdir( oskit_dir_t *dir, const char *name )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    oskit_size_t l ;
    nfs_file_t *d = (nfs_file_t *)dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;

    l = strlen( name ) ;
    if ( l > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_RMDIR ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, name, l ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    return nfs_to_oskit_error( res ) ;
}


static oskit_error_t
fs_dirents_getdirentries( nfs_file_t *dir, oskit_dirents_t **out_de ) ;


static OSKIT_COMDECL
nfs_dir_getdirentries( oskit_dir_t *dir, oskit_u32_t *inout_ofs,
		       oskit_u32_t nentries,
		       struct oskit_dirents **out_dirents )
{
    oskit_dirents_t *de ;
    oskit_error_t err ;

    if ( *inout_ofs != 0 )
    {
        /* We have already returned all the entries. */
        *out_dirents = 0 ;
        return 0 ;
    }

    /* Our dirents object does not take nentries into account. */
    err = fs_dirents_getdirentries( (nfs_file_t *)dir, &de ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    *inout_ofs = 1 ;
    *out_dirents = de ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_dir_mknod(oskit_dir_t *d, const char *name,
		oskit_mode_t mode, oskit_dev_t dev)
{
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_dir_symlink( oskit_dir_t *dir, const char *link_name, const char *dest_name )
{
    oskit_error_t err ;
    oskit_u32_t res ;
    nfs_file_t *d = (nfs_file_t *)dir ;
    oskit_rpc_t *rpc = d->fs->rpc ;
    oskit_size_t lnl ;
    oskit_size_t dnl ;
    nfs_sattr sa ;

    lnl = strlen( link_name ) ;
    if ( lnl > NFS_MAXNAMLEN ) return OSKIT_ENAMETOOLONG ;
    dnl = strlen( dest_name ) ;
    if ( dnl > NFS_MAXPATHLEN ) return OSKIT_ENAMETOOLONG ;

    oskit_rpc_call_init( rpc, NFSPROC_SYMLINK ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;
    oskit_rpc_call_wxdr_opaque( rpc, link_name, lnl ) ;
    oskit_rpc_call_wxdr_opaque( rpc, dest_name, dnl ) ;
    sa.mode = 0777 ;
    sa.uid = 0 ;
    sa.gid = 0 ;
    sa.size = 0 ;
    sa.atime.tv_sec = -1 ;
    sa.atime.tv_usec = -1 ;
    sa.mtime.tv_sec = -1 ;
    sa.mtime.tv_usec = -1 ;
    nfs_call_wxdr_sattr( rpc, &sa ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    return nfs_to_oskit_error( res ) ;
}


/*
 * Create a new "virtual" directory object
 * referring to the same underlying directory as this one,
 * but whose logical parent directory
 * (the directory named by the '..' entry in this directory)
 * is the specified directory node handle.
 *
 * If 'parent' is NULL, lookups of the '..' entry in the
 * virtual directory object will return a reference to
 * the virtual directory object itself; as with physical
 * root directories, virtual root directories will have
 * self-referential '..' entries.
 */
static OSKIT_COMDECL
nfs_dir_reparent( oskit_dir_t *dir, oskit_dir_t *parent, oskit_dir_t **out_dir )
{
    nfs_file_t *d = (nfs_file_t *)dir ;

    if ( parent == d->parent )
    {
        d->ref_counter++ ;
        *out_dir = dir ;
        return 0 ;
    }

    if ( parent ) oskit_dir_addref( parent ) ;
    d->fs->ref_counter++ ;
    *out_dir = (oskit_dir_t *)nfs_file_allocate( d->fs, d->name, &(d->fh),
                                                 &nfs_dir_ops, parent ) ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_file_getfs(oskit_file_t *file, struct oskit_filesystem **out_fs)
{
    nfs_fsys_t *fs = ((nfs_file_t *)file)->fs ;

    fs->ref_counter++ ;
    *out_fs = (oskit_filesystem_t *)fs ;
    return 0 ;
}


static struct oskit_file_ops nfs_file_ops =
{
    nfs_file_query,
    nfs_file_addref,
    nfs_file_release,
    nfs_file_stat,
    nfs_file_setstat,
    no_pathconf,
    nfs_file_sync,
    nfs_file_datasync,
    nfs_file_access,
    nfs_file_readlink,
    nfs_file_open,
    nfs_file_getfs
} ;


typedef OSKIT_COMDECL_U (*dir_addref)(oskit_dir_t *d) ;
typedef OSKIT_COMDECL (*dir_stat)(oskit_dir_t *f, struct oskit_stat *st) ;
typedef OSKIT_COMDECL (*dir_setstat)
    (oskit_dir_t *f, oskit_u32_t mask, const struct oskit_stat *stats) ;
typedef OSKIT_COMDECL
    (*dir_pathconf)(oskit_dir_t *d, oskit_s32_t option, oskit_s32_t *out_val) ;
typedef OSKIT_COMDECL (*dir_sync)(oskit_dir_t *d, oskit_bool_t wait) ;
typedef OSKIT_COMDECL (*dir_access)(oskit_dir_t *d, oskit_amode_t mask) ;
typedef OSKIT_COMDECL (*dir_readlink)
    (oskit_dir_t *d, char *buf, oskit_u32_t len, oskit_u32_t *out_actual) ;
typedef OSKIT_COMDECL (*dir_getfs)
    (oskit_dir_t *f, struct oskit_filesystem **out_fs) ;


static struct oskit_dir_ops nfs_dir_ops =
{
    nfs_dir_query,
    (dir_addref) nfs_file_addref,
    (dir_addref) nfs_file_release,
    (dir_stat) nfs_file_stat,
    (dir_setstat) nfs_file_setstat,
    (dir_pathconf) no_pathconf,
    (dir_sync) nfs_file_sync,
    (dir_sync) nfs_file_datasync,
    (dir_access) nfs_file_access,
    (dir_readlink) nfs_file_readlink,
    nfs_dir_open,
    (dir_getfs) nfs_file_getfs,
    nfs_dir_lookup,
    nfs_dir_create,
    nfs_dir_link,
    nfs_dir_unlink,
    nfs_dir_rename,
    nfs_dir_mkdir,
    nfs_dir_rmdir,
    nfs_dir_getdirentries,
    nfs_dir_mknod,
    nfs_dir_symlink,
    nfs_dir_reparent
} ;


/*******************************************************************************
oskit_filesystem_t interface.
*******************************************************************************/


static OSKIT_COMDECL
nfs_fsys_query( oskit_filesystem_t *fs,
                const oskit_iid_t *iid, void **out_ihandle )
{
    nfs_fsys_t *nfs = (nfs_fsys_t *)fs ;

    if ( memcmp( iid, &oskit_iunknown_iid, sizeof(*iid) ) == 0 ||
         memcmp( iid, &oskit_filesystem_iid, sizeof(*iid) ) == 0)
    {
        *out_ihandle = nfs ;
        nfs->ref_counter++ ;
        return 0 ;
    }

    *out_ihandle = 0 ;
    return OSKIT_E_NOINTERFACE ;
}


static OSKIT_COMDECL_U
nfs_fsys_addref( oskit_filesystem_t *fs )
{
	return ++(((nfs_fsys_t *)fs)->ref_counter) ;
}


static OSKIT_COMDECL_U
nfs_fsys_release( oskit_filesystem_t *fs )
{
    nfs_fsys_t *nfs = (nfs_fsys_t *)fs ;
    nfs_file_t *root ;
    oskit_osenv_mem_t *mem ;

    /* FS hold one reference on FS->ROOT, so this root directory can not be
    released before FS. When we do the last release on FS, we check if the
    root directory has still external references on it */

    if ( --nfs->ref_counter )
        return nfs->ref_counter ;

    /* This is the last release on FS. */
    root = (nfs_file_t *)nfs->root ;

    /* If there are still other references on the root directory, we do not
    release FS. */
    if ( root->ref_counter > 1 )
        return nfs->ref_counter = 1 ;

    if ( root->ref_counter == 1 )
    {
        /* The root directory has to be released. We prevent a loop by setting
        a false ref_counter on NFS. */
        nfs->ref_counter = 2 ;
        nfs_file_release( (oskit_file_t *)(nfs->root) ) ;
    }

    mem = nfs->mem ;
    oskit_rpc_release( nfs->rpc ) ;
    OSKIT_OSENV_MEM_FREE( mem, nfs->hostname,
                          OSENV_NONBLOCKING, strlen(nfs->hostname) + 1 ) ;
    OSKIT_OSENV_MEM_FREE( mem, nfs->dirname,
                          OSENV_NONBLOCKING, strlen(nfs->dirname) + 1 ) ;
    OSKIT_OSENV_MEM_FREE( mem, nfs, OSENV_NONBLOCKING, sizeof( *nfs ) ) ;
    oskit_osenv_mem_release( mem ) ;
    return 0 ;
}


/*
 * Return general information about the file system
 * on which this node resides.
 */
static OSKIT_COMDECL
nfs_fsys_statfs( oskit_filesystem_t *f, oskit_statfs_t *out_stats )
{
    nfs_file_t *d = (nfs_file_t *)(((nfs_fsys_t *)f)->root) ;
    oskit_rpc_t *rpc = ((nfs_fsys_t *)f)->rpc ;
    oskit_error_t err ;
    oskit_u32_t res ;

    oskit_rpc_call_init( rpc, NFSPROC_STATFS ) ;
    oskit_rpc_call_wxdr_data( rpc, d->fh, sizeof(nfs_fhandle) ) ;

    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == NFS_OK )
    {
        /* No error. */
        if ( oskit_rpc_reply_len( rpc ) != 20 )
            return OSKIT_E_FAIL ;

        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        out_stats->bsize = res ;    /* file system block size */
        out_stats->frsize = res ;   /* fundamental file system block size */
        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        out_stats->blocks = res ;   /* total blocks in fs in units of frsize */
        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        out_stats->bfree = res ;    /* free blocks in fs */
        oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        out_stats->bavail = res ;   /* free blocks avail to non-superuser */

        /* XXX Those variables are not specified by NFS. */
        out_stats->favail = 0 ;     /* free file nodes avail to non-superuser */
        out_stats->files = 0 ;      /* total file nodes in file system */
        out_stats->ffree = 0 ;      /* free file nodes in fs */
        out_stats->fsid = 0 ;       /* file system id */
        out_stats->flag = 0 ;       /* mount flags */
        out_stats->namemax = NFS_MAXNAMLEN ;   /* max bytes in a file name */
        return 0 ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }
}


/*
 * Write all data back to permanent storage.
 * If wait is true, this call should not return
 * until all pending data have been completely written;
 * if wait is false, begin the I/O but do not wait
 * to verify that it completes successfully.
 */
static OSKIT_COMDECL
nfs_fsys_sync(oskit_filesystem_t *f, oskit_bool_t wait)
{
    return 0 ;
}


/*
 * Return a reference to the root directory of this file system.
 */
static OSKIT_COMDECL
nfs_fsys_getroot(oskit_filesystem_t *f, oskit_dir_t **out_dir)
{
    nfs_fsys_t *fs = (nfs_fsys_t *)f ;

    nfs_file_addref( (oskit_file_t *)fs->root ) ;
    *out_dir = fs->root ;
    return 0 ;
}


/*
 * Update the flags of this filesystem (eg rdonly -> rdwr).
 * May return OSKIT_E_NOTIMPL if this is not
 * a typical disk-based file system.
 */
static OSKIT_COMDECL
nfs_fsys_remount( oskit_filesystem_t *f, oskit_u32_t flags )
{
    return OSKIT_E_NOTIMPL ;
}


/*
 * Forcibly unmount this filesystem.
 *
 * A normal unmount occurs when all references
 * to the filesystem are released.  This
 * operation forces the filesystem to be
 * unmounted, even if references still exist,
 * and is consequently unsafe.
 *
 * Subsequent attempts to use references to the
 * filesystem or to use references to files within
 * the filesystem may yield undefined results.
 */
static OSKIT_COMDECL
nfs_fsys_unmount( oskit_filesystem_t *f )
{
    /* sure, unmount me */
    return 0 ;
}


static OSKIT_COMDECL
nfs_fsys_lookupi( oskit_filesystem_t *f, oskit_ino_t ino,
                  oskit_file_t **out_file )
{
    return OSKIT_E_NOTIMPL ;
}


static struct oskit_filesystem_ops nfs_fsys_ops =
{
    nfs_fsys_query,
    nfs_fsys_addref,
    nfs_fsys_release,
    nfs_fsys_statfs,
    nfs_fsys_sync,
    nfs_fsys_getroot,
    nfs_fsys_remount,
    nfs_fsys_unmount,
    nfs_fsys_lookupi
} ;


/*******************************************************************************
oskit_absio_t and oskit_bufio_t interfaces.
*******************************************************************************/


/* Get the address of the nfs_file_t containing bufio. */
#define BUFIO2FILE(bufio) ((nfs_file_t *) \
    (((char *)(bufio))-(char *)&(((nfs_file_t *)0)->bufioi)))
#define BUFIO2OSKITFILE(bufio) ((oskit_file_t *)BUFIO2FILE(bufio))


static OSKIT_COMDECL
nfs_bufio_query( oskit_bufio_t *io, const struct oskit_guid *iid,
                 void **out_ihandle )
{
    return nfs_file_query( BUFIO2OSKITFILE(io), iid, out_ihandle ) ;
}


static OSKIT_COMDECL_U
nfs_bufio_addref(oskit_bufio_t *io)
{
    return ++BUFIO2FILE(io)->ref_counter ;
}


static OSKIT_COMDECL_U
nfs_bufio_release(oskit_bufio_t *io)
{
    return nfs_file_release( BUFIO2OSKITFILE( io ) ) ;
}


static OSKIT_COMDECL
nfs_bufio_read( oskit_bufio_t *io, void *buf,
                oskit_off_t offset, oskit_size_t amount,
                oskit_size_t *out_actual )
{
    nfs_file_t *f = BUFIO2FILE(io) ;
    char *p = buf ;
    oskit_size_t sz = 0 ;
    oskit_rpc_t *rpc = f->fs->rpc ;

    if ( offset < 0 ) return OSKIT_EINVAL ;

    /* XXX The operation is not atomic !!! */
    while ( amount != 0 )
    {
        oskit_size_t l ;
        oskit_error_t err ;
        oskit_u32_t res ;

        l = ( amount > 1024 ) ? 1024 : amount ;

        oskit_rpc_call_init( rpc, NFSPROC_READ ) ;
        oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;
        oskit_rpc_call_wxdr_int( rpc, offset ) ;
        oskit_rpc_call_wxdr_int( rpc, l ) ;
        oskit_rpc_call_wxdr_int( rpc, 0 ) ;

        do
        {
            err = oskit_rpc_do_call( rpc ) ;
        }
        while( err == OSKIT_ETIMEDOUT ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        if ( res == NFS_OK )
        {
            char *ptr ;

            /* Skip the file attributes. */
            if ( oskit_rpc_reply_len( rpc ) < sizeof(nfs_fattr) )
                return OSKIT_E_FAIL ;
            oskit_rpc_reply_rxdr_data( rpc, (void**)&ptr, sizeof(nfs_fattr) ) ;

            err = oskit_rpc_reply_rxdr_opaque( rpc, &ptr, &res ) ;
            if ( OSKIT_FAILED( err ) ) return err ;

            /* Should never happen. */
            if ( res > l ) return OSKIT_E_FAIL ;

            memcpy( p, ptr, res ) ;
            amount -= res ;
            sz += res ;
            offset += res ;
            p += res ;

            /* XXX I think this is a good reason to stop the read. */
            if ( res < l ) amount = 0 ;
        }
        else
        {
            /* An error. */
            /* XXX What should I do here ? */
            return nfs_to_oskit_error( res ) ;
        }
    }

    *out_actual = sz ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_bufio_write( oskit_bufio_t *io, const void *buf,
                 oskit_off_t offset, oskit_size_t amount,
                 oskit_size_t *out_actual )
{
    nfs_file_t *f = BUFIO2FILE(io) ;
    const char *p = buf ;
    oskit_size_t sz = 0 ;
    oskit_rpc_t *rpc = f->fs->rpc ;

    if ( offset < 0 ) return OSKIT_EINVAL ;

    /* XXX The operation is not atomic !!! */
    while ( amount != 0 )
    {
        oskit_size_t l ;
        oskit_error_t err ;
        oskit_u32_t res ;

        l = ( amount > 1024 ) ? 1024 : amount ;
        amount -= l ;
        sz += l ;

        oskit_rpc_call_init( rpc, NFSPROC_WRITE ) ;
        oskit_rpc_call_wxdr_data( rpc, f->fh, sizeof(nfs_fhandle) ) ;
        oskit_rpc_call_wxdr_int( rpc, 0 ) ;
        oskit_rpc_call_wxdr_int( rpc, offset ) ;
        oskit_rpc_call_wxdr_int( rpc, 0 ) ;
        oskit_rpc_call_wxdr_opaque( rpc, p, l ) ;
        offset += l ;
        p += l ;

        do
        {
            err = oskit_rpc_do_call( rpc ) ;
        }
        while( err == OSKIT_ETIMEDOUT ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        if ( res == NFS_OK )
        {

            if ( oskit_rpc_reply_len( rpc ) != sizeof(nfs_fattr) )
                return OSKIT_E_FAIL ;
        }
        else
        {
            /* An error. */
            /* XXX What should I do here ? */
            return nfs_to_oskit_error( res ) ;
        }
    }

    *out_actual = sz ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_bufio_get_size( oskit_bufio_t *io, oskit_off_t *out_size )
{
    oskit_error_t err ;
    nfs_fattr fa ;

    err = nfsproc_getattr( BUFIO2FILE(io), &fa ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    *out_size = fa.size ;
    return 0 ;
}


static OSKIT_COMDECL
nfs_bufio_set_size( oskit_bufio_t *io, oskit_off_t new_size )
{
    nfs_sattr sa ;

    sa.mode = -1 ;
    sa.uid = -1 ;
    sa.gid = -1 ;
    sa.size = new_size ;
    sa.atime.tv_sec = -1 ;
    sa.atime.tv_usec = -1 ;
    sa.mtime.tv_sec = -1 ;
    sa.mtime.tv_usec = -1 ;
    return nfsproc_setattr( BUFIO2FILE(io), &sa ) ;
}


static OSKIT_COMDECL
nfs_bufio_map(oskit_bufio_t *io, void **out_addr,
       oskit_off_t offset, oskit_size_t count)
{
    /* Implementation not needed and not interesting. */
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_bufio_unmap(oskit_bufio_t *io, void *addr, 
	oskit_off_t offset, oskit_size_t count)
{
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_bufio_wire(oskit_bufio_t *io, oskit_addr_t *out_phys_addr, 
	oskit_off_t offset, oskit_size_t count)
{
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_bufio_unwire(oskit_bufio_t *io, oskit_addr_t phys_addr, 
		oskit_off_t offset, oskit_size_t count)
{
    return OSKIT_E_NOTIMPL ;
}


static OSKIT_COMDECL
nfs_bufio_copy(oskit_bufio_t *io, oskit_off_t offset, 
	oskit_size_t count, oskit_bufio_t **out_io)
{
    return OSKIT_E_NOTIMPL ;
}


static struct oskit_bufio_ops nfs_bufio_ops =
{
    nfs_bufio_query,
    nfs_bufio_addref,
    nfs_bufio_release,
    (void *)0,			/* slot reserved for getblocksize */
    nfs_bufio_read,
    nfs_bufio_write,
    nfs_bufio_get_size,
    nfs_bufio_set_size,
    nfs_bufio_map,
    nfs_bufio_unmap,
    nfs_bufio_wire,
    nfs_bufio_unwire,
    nfs_bufio_copy
} ;


/*******************************************************************************
The dirents interface.
*******************************************************************************/


typedef struct dirent_lst
{
    struct dirent_lst *next ;
    oskit_u32_t fileid ;
    oskit_u32_t namelen ;   /* Not including the terminal 0. */
    char name[NFS_MAXNAMLEN+1] ;
} dirent_lst_t ;


#define DE_LST_LEN(nlen) ( sizeof(dirent_lst_t) + nlen - NFS_MAXNAMLEN )


typedef struct fs_dirents
{
    oskit_dirents_t direntsi ;
    oskit_u32_t ref_counter ;
    nfs_file_t *dir ;   /* The directory in which we are searching. */
    dirent_lst_t *lst ; /* Start of the list. */
    dirent_lst_t *cur ; /* Pointer in the list. */
    oskit_u32_t len ;   /* Length of the list. */
} fs_dirents_t ;


static struct oskit_dirents_ops fs_dirents_ops ;


/* A reference must be added on DIR before call. */
static oskit_error_t
fs_dirents_getdirentries( nfs_file_t *dir, oskit_dirents_t **out_de )
{
    oskit_u32_t eof = 0 ;
    oskit_u32_t cookie = 0 ;
    oskit_rpc_t *rpc = dir->fs->rpc ;
    oskit_osenv_mem_t *mem = dir->mem ;
    oskit_u32_t len = 0 ;
    fs_dirents_t *de ;
    dirent_lst_t **nxt ;

    de = OSKIT_OSENV_MEM_ALLOC( mem, sizeof(fs_dirents_t),
                                OSENV_NONBLOCKING, 0 ) ;

    if ( de == 0 ) return OSKIT_ENOMEM ;
    de->direntsi.ops = &fs_dirents_ops ;
    de->ref_counter = 1 ;
    de->dir = dir ;
    nxt = &(de->lst) ;

    while( !eof )
    {
        oskit_error_t err ;
        oskit_u32_t res ;

        oskit_rpc_call_init( rpc, NFSPROC_READDIR ) ;
        oskit_rpc_call_wxdr_data( rpc, dir->fh, sizeof(nfs_fhandle) ) ;
        oskit_rpc_call_wxdr_int( rpc, cookie ) ;
        oskit_rpc_call_wxdr_int( rpc, 1024 ) ;

        do
        {
            err = oskit_rpc_do_call( rpc ) ;
        }
        while( err == OSKIT_ETIMEDOUT ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
        if ( OSKIT_FAILED( err ) ) return err ;

        if ( res == NFS_OK )
        {
            err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
            if ( OSKIT_FAILED( err ) ) return err ;

            while( res )
            {
                char *p ;
                oskit_u32_t fileid ;
                dirent_lst_t *de_lst ;

                err = oskit_rpc_reply_rxdr_int( rpc, &fileid ) ;
                if ( OSKIT_FAILED( err ) ) return err ;

                err = oskit_rpc_reply_rxdr_opaque( rpc, &p, &res ) ;
                if ( OSKIT_FAILED( err ) ) return err ;

                de_lst = OSKIT_OSENV_MEM_ALLOC( mem, DE_LST_LEN(res),
                                                OSENV_NONBLOCKING, 0 );
                if ( de_lst == 0 )
                {
                    /* XXX I should free all the allocated memory. */
                    return OSKIT_ENOMEM ;
                }
                memcpy( de_lst->name, p, res ) ;
                de_lst->name[res] = 0 ;
                de_lst->namelen = res ;
                de_lst->fileid = fileid ;

                *nxt = de_lst ;
                nxt = &(de_lst->next) ;
                len++ ;

                err = oskit_rpc_reply_rxdr_int( rpc, &cookie ) ;
                if ( OSKIT_FAILED( err ) ) return err ;

                err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
                if ( OSKIT_FAILED( err ) ) return err ;
            }

            err = oskit_rpc_reply_rxdr_int( rpc, &eof ) ;
            if ( OSKIT_FAILED( err ) ) return err ;
        }
        else
        {
            /* An error. */
            /* XXX What should I do here ? */
            return nfs_to_oskit_error( res ) ;
        }
    }

    *nxt = 0 ;
    de->cur = de->lst ;
    de->len = len ;
    *out_de = (oskit_dirents_t *)de ;
    return 0 ;
}


static OSKIT_COMDECL
fs_dirents_query( oskit_dirents_t *d,
		  const oskit_iid_t *iid, void **out_ihandle )
{
    fs_dirents_t *fsdirents = (fs_dirents_t *)d ;

    if ( memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
         memcmp(iid, &oskit_dirents_iid, sizeof(*iid)) == 0 )
    {
        *out_ihandle = fsdirents ;
        ++fsdirents->ref_counter ;
        return 0 ;
    }

    *out_ihandle = 0 ;
    return OSKIT_E_NOINTERFACE ;
}


static OSKIT_COMDECL_U
fs_dirents_addref( oskit_dirents_t *d )
{
    return ++((fs_dirents_t *)d)->ref_counter ;
}


static OSKIT_COMDECL_U
fs_dirents_release( oskit_dirents_t *d )
{
    fs_dirents_t *de = (fs_dirents_t *) d ;
    nfs_file_t *dir ;
    oskit_osenv_mem_t *mem ;
    dirent_lst_t *l ;

    if ( --de->ref_counter )
        return de->ref_counter ;

    dir = de->dir ;
    mem = dir->mem ;
    l = de->lst ;

    while( l != 0 )
    {
        dirent_lst_t *l2 = l ;

        l = l->next ;
        OSKIT_OSENV_MEM_FREE( mem, l2, OSENV_NONBLOCKING,
                              DE_LST_LEN(l2->namelen) ) ;
    }

    OSKIT_OSENV_MEM_FREE( mem, de, OSENV_NONBLOCKING,
                          sizeof(fs_dirents_t) ) ;
    nfs_file_release( (oskit_file_t *)dir ) ;

    return 0 ;
}


static OSKIT_COMDECL
fs_dirents_getcount( oskit_dirents_t *d, int *out_count )
{
    fs_dirents_t *de = (fs_dirents_t *) d ;

    *out_count = de->len ;

    return 0 ;
}


static OSKIT_COMDECL
fs_dirents_getnext( oskit_dirents_t *d, oskit_dirent_t *dirent )
{
    fs_dirents_t *de = (fs_dirents_t *) d ;
    dirent_lst_t *l = de->cur ;

    if ( l == 0 ) return OSKIT_EWOULDBLOCK ;

    if ( dirent->namelen < l->namelen + 1 )
        return OSKIT_ENAMETOOLONG ;

    dirent->ino = l->fileid ;
    dirent->namelen = l->namelen ;
    strcpy( dirent->name, l->name ) ;

    de->cur = l->next ;
    return 0 ;
}


static OSKIT_COMDECL
fs_dirents_rewind( oskit_dirents_t *d )
{
    fs_dirents_t *de = (fs_dirents_t *) d ;

    de->cur = de->lst ;
    return 0 ;
}


static struct oskit_dirents_ops fs_dirents_ops =
{
    fs_dirents_query,
    fs_dirents_addref,
    fs_dirents_release,
    fs_dirents_getcount,
    fs_dirents_getnext,
    fs_dirents_rewind
};


/*******************************************************************************
Initialization of an NFS filesystem.
*******************************************************************************/


static oskit_error_t
mntproc_mnt( nfs_fsys_t *fs )
{
    oskit_rpc_t *rpc = fs->rpc ;
    oskit_error_t err ;
    oskit_u32_t res ;
    nfs_fhandle *d ;
    nfs_file_t *file ;

    err = oskit_rpc_connect( rpc, fs->hostname, MOUNT_PROG, MOUNT_VERSION ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    oskit_rpc_call_init( rpc, MNTPROC_MNT ) ;
    oskit_rpc_call_wxdr_string( rpc, fs->dirname ) ;
    do
    {
        err = oskit_rpc_do_call( rpc ) ;
    }
    while( err == OSKIT_ETIMEDOUT ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    err = oskit_rpc_reply_rxdr_int( rpc, &res ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    if ( res == 0 )
    {
        /* No error. */
        err = oskit_rpc_reply_rxdr_data( rpc, (void**)&d, sizeof(nfs_fhandle) ) ;
        if ( OSKIT_FAILED( err ) ) return err ;
    }
    else
    {
        /* An error. */
        return nfs_to_oskit_error( res ) ;
    }

    /* d points to the nfs_fhandle. */
    err = oskit_rpc_connect( rpc, fs->hostname, NFS_PROG, NFS_VERSION ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    /* We have a circular reference between FS and FS->ROOT. We set the two
    reference counters to one. nfs_fsys_release() takes it into account. */
    file = nfs_file_allocate( fs, "nfs root directory", d, &nfs_dir_ops, 0 ) ;
    if ( file == 0 ) return OSKIT_ENOMEM ;
    fs->root = (oskit_dir_t *)file ;

    return 0 ;
}


/* Initialize an NFS filesystem.
HOSTNAME: Name of the NFS server,
DIRNAME: Directory on the NFS server,
LOCALHOST: Name of the local host,
SHORT_TIMEOUT: Time in ms before retrying a call,
LONG_TIMEOUT: Time in ms before aborting a call,
 */
oskit_error_t
oskit_nfs_init( oskit_osenv_mem_t *mem,
                oskit_osenv_timer_t *timer,
                oskit_socket_factory_t *sf,
                const char *hostname,
                const char *dirname,
                const char *localhost,
                oskit_u32_t short_timeout,
                oskit_u32_t long_timeout,
                oskit_filesystem_t **out_fs )
{
    nfs_fsys_t *fs ;
    char *h, *d ;
    oskit_size_t hl, dl ;
    oskit_rpc_t *rpc ;
    oskit_error_t err ;

    /* Initialize an RPC object. */
    oskit_osenv_mem_addref( mem ) ;
    err = oskit_rpc_init( mem, timer, sf, localhost,
                          short_timeout, long_timeout, &rpc ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    /* Get memory for the NFS object. */
    hl = strlen( hostname ) + 1 ;
    dl = strlen( dirname ) + 1 ;
    fs = OSKIT_OSENV_MEM_ALLOC( mem, sizeof(*fs), OSENV_NONBLOCKING, 0 );
    h = OSKIT_OSENV_MEM_ALLOC( mem, hl, OSENV_NONBLOCKING, 0);
    d = OSKIT_OSENV_MEM_ALLOC( mem, dl, OSENV_NONBLOCKING, 0);
    if ( (!fs) || (!d) || (!h) )
    {
        oskit_osenv_mem_release( mem ) ;
        oskit_rpc_release( rpc ) ;
        return OSKIT_E_FAIL;
    }
    memcpy( h, hostname, hl ) ;
    memcpy( d, dirname, dl ) ;

    /* Initialize the NFS object. */
    fs->fsysi.ops = &nfs_fsys_ops ;
    fs->hostname = h ;
    fs->dirname = d ;
    fs->ref_counter = 1 ;
    fs->mem = mem ;
    fs->rpc = rpc ;

    err = mntproc_mnt( fs ) ;
    if ( OSKIT_FAILED( err ) )
    {
        oskit_rpc_release( rpc ) ;
        OSKIT_OSENV_MEM_FREE( mem, h, OSENV_NONBLOCKING, hl ) ;
        OSKIT_OSENV_MEM_FREE( mem, d, OSENV_NONBLOCKING, dl ) ;
        OSKIT_OSENV_MEM_FREE( mem, fs, OSENV_NONBLOCKING, sizeof(*fs) ) ;
        oskit_osenv_mem_release( mem ) ;
        return err ;
    }

    *out_fs = (oskit_filesystem_t *)fs ;
    return 0 ;
}
